---------------------------------------------------------
-- Carry-save multiplier for unsigned operands (base 2) 
-- This multiplier is from the book [Deschamps06]
-- For an complete description of this module, see this reference.
-- This module is fully generic
----------------------------------------------------------		
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

entity carry_save_mult is
   generic(NBITS :natural := 18;
	        MBITS :natural := 18);
    Port ( X : in std_logic_vector(NBITS-1 downto 0);
           Y : in std_logic_vector(MBITS-1 downto 0);
           P : out std_logic_vector(MBITS+NBITS-1 downto 0));
end carry_save_mult;

architecture Behavioral of carry_save_mult is

component basic_mul_cell
    Port ( X_i : in std_logic;
           Y_j : in std_logic;
           Cin : in std_logic; 
           Pin : in std_logic;
           Cout : out std_logic;
           Pout : out std_logic
			  );
end component;
type conections is array (0 to NBITS) of STD_LOGIC_VECTOR (MBITS downto 0);
Signal Cin, Pin, Cout, Pout: conections;
begin

Pin(0) <= (others => '0');
Cin(0) <= (others => '0');

ext_loop: for i in 0 to NBITS-1 generate
  int_loop: for j in 0 to MBITS-1 generate
	 intern_a:  basic_mul_cell port map(
		           X_i => X(i), Y_j => Y(j),
		           Cin => Cin(i)(j), Pin => Pin(i)(j),
		           Cout => Cout(i)(j),  Pout => Pout(i)(j) );
				Cin(i+1)(j) <= Cout(i)(j);
				j_0: if J = 0 generate P(i) <= Pout(i)(j); end generate;
				jn0: if J > 0 generate Pin(i+1)(j-1) <= Pout(i)(j); end generate;
	 end generate;
	 Pin(i+1)(MBITS-1) <= '0';
end generate;

P(MBITS+NBITS-1 downto NBITS) <= Pin(NBITS)(MBITS-1 downto 0) + Cin(NBITS)(MBITS-1 downto 0);

end Behavioral;

---------------------------------------------------------
-- Basic multiplier cell 
-- 
----------------------------------------------------------		
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;
--library UNISIM;
--use UNISIM.VComponents.all;

entity basic_mul_cell is
    Port ( X_i : in std_logic;
           Y_j : in std_logic;
           Cin : in std_logic; 
           Pin : in std_logic;
           Cout : out std_logic;
           Pout : out std_logic
			  );
end basic_mul_cell;

architecture Behavioral of basic_mul_cell is
signal int_and: std_logic;
begin
  int_and <= X_i and Y_j;	
  Cout <= (Cin and Pin) or (Cin and int_and) or (Pin and int_and);
  Pout <= Cin xor int_and xor Pin;
end Behavioral;

------------------------------------------------------------------------
---- VHDL Test Bench for Ripple carry multiplier
----
---- Notes: 
------------------------------------------------------------------------
--LIBRARY  IEEE;
--USE IEEE.std_logic_1164.all;
--USE IEEE.std_logic_arith.all;
--USE IEEE.std_logic_unsigned.all;
--USE work.mypackage.all;
--
--LIBRARY ieee;
--USE IEEE.STD_LOGIC_TEXTIO.ALL;
--USE STD.TEXTIO.ALL;
--
--ENTITY test_exhaustive IS
--END test_exhaustive;
--
--ARCHITECTURE behavioural OF test_exhaustive IS 
--constant DELAY: time := 100 ns;
--FILE RESULTS: TEXT OPEN WRITE_MODE IS "results.txt";
--
--	COMPONENT carry_save_mult
--	PORT(
--		X : IN std_logic_vector(NBITS-1 downto 0);
--		Y : IN std_logic_vector(MBITS-1 downto 0);
--		P : OUT std_logic_vector(MBITS+NBITS-1 downto 0)
--		);
--	END COMPONENT;
--
--	SIGNAL x :  std_logic_vector(NBITS-1 downto 0);
--	SIGNAL y :  std_logic_vector(MBITS-1 downto 0);
--	SIGNAL p :  std_logic_vector(MBITS+NBITS-1 downto 0);
--
--BEGIN
--
--	uut: carry_save_mult PORT MAP(X => x, Y => y, P => p);
--
--   tb_test : PROCESS
--	VARIABLE TX_LOC : LINE;
--	VARIABLE TX_STR : String(1 to 4096);
--   BEGIN
--		for I in 0 to 2**NBITS-1 loop
--			for J in 0 to 2**MBITS-1 loop
--				x <= CONV_STD_LOGIC_VECTOR (I, NBITS);
--				y <= CONV_STD_LOGIC_VECTOR (J, MBITS);
--				WAIT FOR DELAY/2;
--				IF ( I*J /= CONV_INTEGER(P)) THEN 
--					write(TX_LOC,string'("ERROR!!! X=")); write(TX_LOC, X);
--					write(TX_LOC,string'(" Y=")); write(TX_LOC, Y);
--					write(TX_LOC,string'(" P=")); write(TX_LOC, P);
--					write(TX_LOC, string'(" "));
--					TX_STR(TX_LOC.all'range) := TX_LOC.all;
--					writeline(results, TX_LOC);
--					Deallocate(TX_LOC);
--					ASSERT (FALSE) REPORT TX_STR SEVERITY ERROR;
--				END IF;
--				WAIT FOR DELAY/2;
--    		end loop;
--   	end loop;
--		ASSERT (FALSE) REPORT
--		"Simulation successful (not a failure).  No problems detected. "
--		SEVERITY FAILURE;
--   END PROCESS;
--
--END;
