---------------------------------------------------------------
-- Testbench for testing the big top-level with the parallell color prosessing architecture.
-- The input is an txt-file named pixels.txt which is generated by the cut_picture-function in Matlab.
-- The output is the txt-file picture_output.txt which may be converted to an bitmap 
-- with the Matlab function read_vhdl_output.
--
-- The parameters f_inn, m_inn, a and c may be adjusted to tune the picture.
--
-- The picture must be processed twice to first calculate the compression-parameters before the real compression.
-- To test the estimation of the parameters you may use one picture to calculate the parameters, then an other
-- picture to compress. 
---------------------------------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;
use IEEE.math_real.all;

library STD;
use STD.textio.all;

---------------------------------------------------------------
-- Entity
---------------------------------------------------------------
entity tb_big is
end entity tb_big;

---------------------------------------------------------------
-- Architecture
---------------------------------------------------------------
architecture behavior of tb_big is

-- The design is not entirely generic, so this constant should be 18    
    constant NBITS : natural := 18;

-- Defining the DUT    
component top_level_big is
    port(
       clk         : in   std_logic;
       reset_in    : in   std_logic;
       go          : in   std_logic;
       
       R           : in   std_logic_vector (18-1 downto 0);
       G           : in   std_logic_vector (18-1 downto 0);
       B           : in   std_logic_vector (18-1 downto 0);
       
       f_user_in   : in   std_logic_vector (18-1 downto 0);
       cntrl_f     : in   std_logic;
       m_in        : in   std_logic_vector (12 downto 0);
       cntrl_m     : in   std_logic;
       c           : in   std_logic_vector (2 downto 0);
       a           : in   std_logic_vector (2 downto 0);
       
       output_R    : out  std_logic_vector (7 downto 0);
       output_G    : out  std_logic_vector (7 downto 0);
       output_B    : out  std_logic_vector (7 downto 0)
       );
end component top_level_big;

    signal R               : std_logic_vector(NBITS-1 downto 0) := (others=>'0');    
    signal G               : std_logic_vector(NBITS-1 downto 0) := (others=>'0');    
    signal B               : std_logic_vector(NBITS-1 downto 0) := (others=>'0');    
    signal output_R        : std_logic_vector(8-1 downto 0);    
    signal output_G        : std_logic_vector(8-1 downto 0);    
    signal output_B        : std_logic_vector(8-1 downto 0);    
    
    signal a               : std_logic_vector(2 downto 0)       := (others=>'0');    
    signal c               : std_logic_vector(2 downto 0)       := (others=>'0');
    signal m               : std_logic_vector(12 downto 0)      := (others=>'0');    
    signal f_user          : std_logic_vector(NBITS-1 downto 0) := (others=>'0');        
    signal cntrl_f         : std_logic                          := '0';
    signal cntrl_m         : std_logic                          := '0';
    signal clk             : std_logic                          := '0';    
    signal reset           : std_logic                          := '0';    
    signal go              : std_logic                          := '0'; 
    signal valid_output    : std_logic                          := '0';
    signal finished        : std_logic                          := '0';
    
    file   my_input_file   : text  open read_mode is "pixels.txt" ;
    file   my_input_file_2 : text  open read_mode is "pixels.txt" ; 
    file   picture_output  : text  open write_mode is "picture_output.txt" ;   

begin

-- Initializing the DUT
   DUT   :   top_level_big   port map(clk, reset, go, R, G, B, f_user, cntrl_f, m, cntrl_m, c, a, output_R, output_G, output_B );

-- Clock generation
clocking: process
begin
    clk <= not clk;
    wait for 5 ns;
end process clocking;

-- The test    
test: process
   variable L        : line;     -- read lines from file to L            
   variable value    : integer;
   variable GOOD     : boolean;
   variable a_loop_i : integer;
   variable b_loop_i : integer;
   variable c_loop_i : integer;
begin
    
    wait for 2 ns;
-- Setting the input parameters. Adjust these to tune the compression.
    cntrl_f    <= '1';
    f_user     <= "001010000000000000";   -- Intensity          -- Logarithmic number  
    c          <= "000";                  -- Chromatic adaption -- 000
    a          <= "111";
    cntrl_m    <= '1';                    -- Light adaption     -- 111
    m          <= "0000000000000";        -- Contrast           -- 0
    
-- Resets the circuit    
    reset      <= '1';
    wait for 20 ns;
    reset      <= '0';
    wait for 20 ns;
    go         <= '1';
    
-- Test loop. my_input_file is the input to the circuit. The first run is for computing the parameters.
   for a_loop in 1 to 480 loop      
      a_loop_i := a_loop;
      readline(my_input_file, L);
      for b_loop in 1 to 640 loop         
         b_loop_i := b_loop;
         for c_loop in 0 to 2 loop
                c_loop_i := c_loop;
                read(L, value, GOOD);
                case c_loop is
                    when 0 =>
                      R     <= conv_std_logic_vector(value, NBITS);
                    when 1 =>
                      G     <= conv_std_logic_vector(value, NBITS);
                    when 2 =>
                      B     <= conv_std_logic_vector(value, NBITS);
                    when others => null;                    
                 end case;                                 
         end loop;
         wait for 10 ns;
      end loop;
   end loop;   
   
-- Test loop. my_input_file_2 is the input to the circuit. The second run is for compressing the input picture.
-- my_input_file and my_input_file_2 may be slightly different for testing the quality of the parameter-estimation.
   wait for 80 ns;   
   for a_loop in 1 to 480 loop
      a_loop_i := a_loop;
      readline(my_input_file_2, L);
      for b_loop in 1 to 640 loop
         if(b_loop = 4) then
             valid_output <= '1';
         end if;
         b_loop_i := b_loop;
         for c_loop in 0 to 2 loop
                c_loop_i := c_loop;
                read(L, value, GOOD);
                case c_loop is
                    when 0 =>
                      R     <= conv_std_logic_vector(value, NBITS);
                    when 1 =>
                      G     <= conv_std_logic_vector(value, NBITS);
                    when 2 =>
                      B     <= conv_std_logic_vector(value, NBITS);
                    when others => null;                    
                 end case;                
         end loop;
         wait for 10 ns;
      end loop;
   end loop;
-- valid_output is for logging the output. The output is slightly delayed because of the pipelining.   
      wait for 30 ns;
   valid_output <= '0';
-- The finished signals tells the logging-prosess to write to the the log-file.      
   wait for 10 ns;   
   finished   <= '1';
   wait for 10 ns;
   finished   <= '0';
-- End the test.   
   assert(false) severity error;
end process test;

-- Process that logs the output (compressed picture).
logging_ans: process(clk)
    variable L_R      : line;
    variable L_G      : line;
    variable L_B      : line;
    variable width    : integer := 10;
    variable int_write: integer;
    variable empty_ch : character    := ' ';
    variable counter  : integer  := 0;
begin
    if(rising_edge(clk)) then
-- If the output is valid, then log the outputs to the temporary lines.        
       if(valid_output = '1') then
           counter := counter + 1;
           int_write := (conv_integer(output_R));
           write(L_R, empty_ch);
           write(L_R, int_write);           
           int_write := conv_integer(output_G);
           write(L_G, empty_ch);
           write(L_G, int_write);
           int_write := conv_integer(output_B);
           write(L_B, empty_ch);
           write(L_B, int_write);           
       end if;
-- If the test is finished, then write the temporary lines to the file.       
       if(finished = '1') then
           writeline(picture_output, L_R);
           writeline(picture_output, L_G);
           writeline(picture_output, L_B);
       end if;
    end if;
end process logging_ans;

-- Process that calculates the error ideal parameters and saves the values for comparison with the estimated parameters in the simulation.
-- This process is used for debug-purposes only, since the compressed picture is saved in the logging-process.
-- This process is mainly an debug-testbench for the parameter-module. 
check_ans: process(R, G, B, a, f_user, c, m, valid_output, output_R, output_G, output_B, clk)
   variable real_R            : real;
   variable real_G            : real;   
   variable real_B            : real;   
   variable real_Lum          : real;
   variable real_a            : real;
   variable real_f            : real;
   variable real_c            : real;   
   variable real_m            : real;
   variable real_output_R     : real;
   variable real_output_G     : real;
   variable real_output_B     : real;
   variable vhdl_output_R     : real;
   variable vhdl_output_G     : real;
   variable vhdl_output_B     : real;
   
   variable real_lum_sum      : real; 
   variable real_lum_avg      : real; 
   variable real_Llum_sum     : real; 
   variable real_Llum_avg     : real; 
   variable real_sum_R        : real; 
   variable real_avg_R        : real; 
   variable real_sum_G        : real; 
   variable real_avg_G        : real; 
   variable real_sum_B        : real; 
   variable real_avg_B        : real; 
   variable real_counter      : real;
   variable real_reset        : boolean;
   variable avg_r             : real;
   variable avg_g             : real;
   variable avg_b             : real;
   variable avg_lum           : real;
   variable avg_Llum          : real;
   variable real_avg_Llum     : real;
   variable real_avg_exp_lum  : real;
   variable real_Llum_exp     : real;
   variable real_Llum_max     : real;
   variable real_Llum_min     : real;
   variable temp_max          : real;  
   variable temp_min          : real;
   variable real_k            : real;    
   variable temp_k            : real; 
    
begin
-- Calculating the real value of the inputs and outputs
   if(valid_output = '1') then
      vhdl_output_R  := real(conv_integer(output_R));
      vhdl_output_G  := real(conv_integer(output_G));
      vhdl_output_B  := real(conv_integer(output_B));
   end if;       
   real_R         := real(conv_integer(R));
   real_G         := real(conv_integer(G));
   real_B         := real(conv_integer(B));
   real_a         := real(conv_integer(a)) / real(7);
   real_c         := real(conv_integer(c)) / real(7);
   real_f         := real(conv_integer(f_user)) / real(8192);
   real_m         := real(conv_integer(m)) / real(8192);
-- Calculating the luminance
   real_Lum       := real(0.2125)*real_R + real(0.7154)*real_G + real(0.0721)*real_B;
-- Calculates the averages, maximums and minimums while keeping track of the reset as well.      
   if(rising_edge(clk)) then
       if(real_reset = true) then
           real_sum_R   := real(0);
           real_sum_G   := real(0);
           real_sum_B   := real(0);
           real_counter := real(0);
           real_Llum_sum:= real(0);           
           real_Llum_max:= real(0);           
           real_Llum_min:= real(9999999);           
       elsif(go='1')     then      
           real_sum_R   := real_sum_R + real_R;
           real_sum_G   := real_sum_G + real_G;
           real_sum_B   := real_sum_B + real_B;
           real_counter := real_counter + real(1);
           if(real_Lum > real(0)) then
              real_Llum_sum:= real_Llum_sum + log2(real_Lum);
              if(log2(real_Lum)<real_Llum_min) then
                 real_Llum_min := log2(real_Lum);
                 temp_min      := real_Llum_min * real(8192);
              end if;
              if(log2(real_Lum)>real_Llum_max) then
                 real_Llum_max := log2(real_Lum);
                 temp_max      := real_Llum_max * real(8192);
              end if;
           end if;
       end if;
   end if;   
-- Calulates the averages.  
   real_Llum_avg  := real_Llum_sum/ real(307200);
   real_lum_avg   := real_lum_sum/ real(307200);
   real_avg_R     := real_sum_R/ real(307200);
   real_avg_G     := real_sum_G/ real(307200);
   real_avg_B     := real_sum_B/ real(307200);   
   if(avg_Llum>real(0)) then
      real_avg_Llum  := avg_Llum*real(8192);
   end if;

-- Calculating and saving the parameters from one frame to another frame.   
   if(real_counter = real(307207) OR reset = '1') then
       avg_r         := real_avg_R;
       avg_g         := real_avg_G;
       avg_b         := real_avg_B;
       avg_lum       := real_lum_avg;
       avg_Llum      := real_Llum_avg;
       real_reset    := true;
       real_Llum_exp := real_avg_exp_lum;
       if((real_Llum_max - real_Llum_min) > real(0)) then
          real_k := (real_Llum_max-avg_Llum)/ (real_Llum_max - real_Llum_min);
          temp_k := real_k * real(256);
        end if;
   else
       real_reset := false;
   end if;

end process check_ans;

end architecture behavior; 