---------------------------------------
-- The control-module for the small architecture.
---------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;

---------------------------------------
-- Defining the entity
---------------------------------------
entity control_small is
    port (
          clk         : in  std_logic;
          reset       : in  std_logic;
          go          : in  std_logic;          
          reset_out   : out std_logic;                              
    			 R_enable    : out std_logic;
          G_enable    : out std_logic;
          B_enable    : out std_logic;
    			 Lum_enable  : out std_logic;                  
          comp_phase  : out std_logic;
          cntrl_RGB   : out std_logic_vector(1 downto 0)
          );
end entity control_small;

---------------------------------------
-- Architecture
---------------------------------------
architecture behavior of control_small is   

-- Defining states for the internal state-machine
    type super_state is (idle, compressing, calculating_parameters);
    
    constant NR_Pixels         : natural := 3*307200;---1;   -- 
    constant NR_param_steps    : natural := 8-1;        -- 
    
    signal shift_state         : std_logic;
    signal cout_1              : std_logic;
    signal cout_2              : std_logic;
    signal cout_3              : std_logic;
    
    signal fsm                 : super_state   := idle;
    signal next_fsm            : super_state   := idle;
    
    signal counter             : natural := 0;
    
    signal cntrl_RGB_internal  : std_logic_vector(1 downto 0);
    
begin  -- behavior    
    
    cntrl_RGB    <= cntrl_RGB_internal;

-- Finite State Machine       
process(clk)
begin
    if(rising_edge(clk)) then
        if(reset = '1') then
           fsm            <=   idle;
        else
           fsm            <=   next_fsm;
        end if;               
    end if;
end process;

-- Keeping track of which color to process.
process(clk)
begin
    if(rising_edge(clk)) then
        if(next_fsm = compressing OR fsm=compressing OR (counter=NR_param_steps AND go= '1') OR ((counter < 5) AND fsm = calculating_parameters)) then
            if(cntrl_RGB_internal = "11") then
                cntrl_RGB_internal <= "01";
            else
                cntrl_RGB_internal <= cntrl_RGB_internal + 1;
            end if;
        else
           cntrl_RGB_internal <= "00";
        end if;            
    end if;
end process;   

-- Keeping track of when to shift state
keep_count: process(clk)
begin
    if(rising_edge(clk)) then
       if(go='1') then
           if((counter = NR_Pixels AND fsm = compressing) OR (counter = NR_param_steps AND fsm = calculating_parameters)) then
               shift_state <= '1';
               counter <= 0;   --(others => 0);
           else
               shift_state <= '0';
               counter <= counter +1;
           end if;
       else
               shift_state <= '0';
               counter <= 0;
       end if;       
    end if;
end process keep_count;

-- Setting the output-signals of according to the fsm
process(go, fsm, shift_state, cntrl_RGB_internal)
begin
        case(fsm) is
            -- Idle state-just waiting for the 'go'-signal
          when idle =>              
             R_enable       <= '0';
             G_enable       <= '0';
             B_enable       <= '0';
             Lum_enable     <= '0';
             reset_out      <= '1';
             comp_phase     <= '0';
             if(go='1') then
                 next_fsm <= compressing;
             else
                 next_fsm <= idle;
             end if;
             
             -- compressing state- compressing the pixels one frame at a time
          when compressing =>
              case cntrl_RGB_internal is 
              -- Compress Red
                 when "01" => 
                   R_enable       <= '1';
                   G_enable       <= '0';
                   B_enable       <= '0';
                   Lum_enable     <= '1';
              -- Compress Green
                 when "10" => 
                   R_enable       <= '0';
                   G_enable       <= '1';
                   B_enable       <= '0';
                   Lum_enable     <= '0';
              -- Compress Blue
                 when "11" => 
                   R_enable       <= '0';
                   G_enable       <= '0';
                   B_enable       <= '1';
                   Lum_enable     <= '0';
              -- Idle
                 when others => 
                   R_enable       <= '0';
                   G_enable       <= '0';
                   B_enable       <= '0';
                   Lum_enable     <= '0';
               end case;
             reset_out      <= '0';            
             if(shift_state = '1') then
                 comp_phase      <= '0';
                 if(go='1') then
                    next_fsm     <= calculating_parameters;
                 else                    
                    next_fsm     <= idle;
                 end if;
             else
                  comp_phase      <= '1';
             end if;

             -- calculating_parameters - calculating next frames parameters from the last frames values
          when calculating_parameters =>
             R_enable       <= '0';
             G_enable       <= '0';
             B_enable       <= '0';
             Lum_enable     <= '0';
             reset_out      <= '0';
             if(shift_state = '1') then
                 comp_phase     <= '1';
                 if(go='1') then
                    next_fsm <= compressing;
                    R_enable       <= '1';
                    G_enable       <= '0';
                    B_enable       <= '0';
                    Lum_enable     <= '1';
                 else
                    next_fsm <= idle;
                 end if;
             else
                 comp_phase     <= '0';
             end if;
             
          when others =>   next_fsm <= idle;
          end case;                
end process;

end architecture behavior;  -- of 