---------------------------------------------------------------
-- Testbench for testing the smaller top-level with the serial color prosessing architecture.
-- The input is an txt-file named pixels.txt which is generated by the cut_picture-function in Matlab.
-- The output is the txt-file picture_output.txt which may be converted to an bitmap 
-- with the Matlab function read_vhdl_output.
--
-- The parameters f_inn, m_inn, a and c may be adjusted to tune the picture.
--
-- The picture must be processed twice to first calculate the compression-parameters before the real compression.
-- To test the estimation of the parameters you may use one picture to calculate the parameters, then an other
-- picture to compress. 
---------------------------------------------------------------

library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;
use IEEE.math_real.all;

library STD;
use STD.textio.all;

---------------------------------------------------------------
-- Entity
---------------------------------------------------------------

entity tb_small is
end entity tb_small;

---------------------------------------------------------------
-- Architecture
---------------------------------------------------------------

architecture behavior of tb_small is

-- The design is not entirely generic, so this constant should be 18     
    constant NBITS : natural := 18;

-- Defining the DUT       
component top_level_small is
    port(
       clk         : in   std_logic;
       reset_in    : in   std_logic;
       go          : in   std_logic;
       
       R           : in   std_logic_vector (18-1 downto 0);
       G           : in   std_logic_vector (18-1 downto 0);
       B           : in   std_logic_vector (18-1 downto 0);
       
       f_user_in   : in   std_logic_vector (18-1 downto 0);
       cntrl_f     : in   std_logic;
       m_in        : in   std_logic_vector (12 downto 0);
       cntrl_m     : in   std_logic;
       c           : in   std_logic_vector (2 downto 0);
       a           : in   std_logic_vector (2 downto 0);
       
       output_R    : out  std_logic_vector (7 downto 0);
       output_G    : out  std_logic_vector (7 downto 0);
       output_B    : out  std_logic_vector (7 downto 0)
       );
end component top_level_small;

    signal R               : std_logic_vector(NBITS-1 downto 0) := (others=>'0');    
    signal G               : std_logic_vector(NBITS-1 downto 0) := (others=>'0');    
    signal B               : std_logic_vector(NBITS-1 downto 0) := (others=>'0');    
    signal output_R        : std_logic_vector(8-1 downto 0);    
    signal output_G        : std_logic_vector(8-1 downto 0);    
    signal output_B        : std_logic_vector(8-1 downto 0);    
    
    signal a               : std_logic_vector(2 downto 0)       := (others=>'0');    
    signal c               : std_logic_vector(2 downto 0)       := (others=>'0');
    signal m               : std_logic_vector(12 downto 0)      := (others=>'0');    
    signal f_user          : std_logic_vector(NBITS-1 downto 0) := (others=>'0');        
    signal cntrl_f         : std_logic                          := '0';
    signal cntrl_m         : std_logic                          := '0';
    signal clk             : std_logic                          := '0';    
    signal reset           : std_logic                          := '0';    
    signal go              : std_logic                          := '0'; 
    signal valid_output    : std_logic                          := '0';
    signal finished        : std_logic                          := '0';
    
    file   my_input_file   : text  open read_mode is "pixels.txt" ;
    file   my_input_file_2 : text  open read_mode is "pixels.txt" ; 
    file   picture_output  : text  open write_mode is "picture_output_small.txt" ;   

begin  
    
-- Initializing the DUT
   DUT   :   top_level_small   port map(clk, reset, go, R, G, B, f_user, cntrl_f, m, cntrl_m, c, a, output_R, output_G, output_B );

-- Clock generation
clocking: process
begin
    clk <= not clk;
    wait for 5 ns;
end process clocking;

-- The test       
test: process
   variable L        : line;              -- read lines from file to L  
   variable value    : integer;
   variable GOOD     : boolean;
   variable a_loop_i : integer;
   variable b_loop_i : integer;
   variable c_loop_i : integer;
begin
--    readline(my_input_file, L);
    
    wait for 2 ns;
    
-- Setting the input parameters. Adjust these to tune the compression.
    cntrl_f    <= '1';
    f_user     <= "001010000000000000";   -- Intensity          -- Logarithmic number  
    c          <= "000";                  -- Chromatic adaption -- 000
    a          <= "111";                  -- Light adaption     -- 111
    cntrl_m    <= '1';                    
    m          <= "0000000000000";        -- Contrast           -- 0
    
-- Resets the circuit    
    reset      <= '1';
    wait for 20 ns;
    reset      <= '0';
    wait for 20 ns;
    go         <= '1';                

-- Test loop. my_input_file is the input to the circuit. The first run is for computing the parameters.    
   for a_loop in 1 to 480 loop      
      a_loop_i := a_loop;
      readline(my_input_file, L);
      for b_loop in 1 to 640 loop         
         b_loop_i := b_loop;
         for c_loop in 0 to 2 loop
                c_loop_i := c_loop;
                read(L, value, GOOD);
                case c_loop is
                    when 0 =>
                      R     <= conv_std_logic_vector(value, NBITS);
                    when 1 =>
                      G     <= conv_std_logic_vector(value, NBITS);
                    when 2 =>
                      B     <= conv_std_logic_vector(value, NBITS);
                    when others => null;                    
                 end case;                                                  
         end loop;
         wait for 30 ns;
      end loop;
   end loop;   
   
   wait for 90 ns;
   
-- Test loop. my_input_file_2 is the input to the circuit. The second run is for compressing the input picture.
-- my_input_file and my_input_file_2 may be slightly different for testing the quality of the parameter-estimation.   
   for a_loop in 1 to 480 loop
      a_loop_i := a_loop;
      readline(my_input_file_2, L);
      for b_loop in 1 to 640 loop
         b_loop_i := b_loop;
         for c_loop in 0 to 2 loop
                c_loop_i := c_loop;
                read(L, value, GOOD);
                case c_loop is
                    when 0 =>
                      R     <= conv_std_logic_vector(value, NBITS);
                    when 1 =>
                      G     <= conv_std_logic_vector(value, NBITS);
                    when 2 =>
                      B     <= conv_std_logic_vector(value, NBITS);
                    when others => null;                    
                 end case;                
         end loop;
         wait for 20 ns;
         if(b_loop = 2) then
             valid_output <= '1';
         end if;
         wait for 10 ns;
      end loop;
   end loop;
-- valid_output is for logging the output. The output is slightly delayed because of the pipelining and serial color-compression.      
      wait for 50 ns;
   valid_output <= '0';
   wait for 10 ns;
-- The finished signals tells the logging-prosess to write to the the log-file.      
   finished   <= '1';
   wait for 10 ns;
   finished   <= '0';   
   wait for 100 ns;
-- End the test.   
   assert(false) severity error;
end process test;

-- Process that logs the output (compressed picture).
logging_ans: process(clk)
    variable L_R      : line;
    variable width    : integer := 10;
    variable int_write: integer;
    variable empty_ch : character    := ' ';
    variable counter  : integer  := 0;
begin
    if(rising_edge(clk)) then
-- If the output is valid, then log the outputs to the temporary lines.  
-- Only the output_R is valid in this architecture.        
       if(valid_output = '1') then
           counter := counter + 1;
           int_write := (conv_integer(output_R));
           write(L_R, empty_ch);
           write(L_R, int_write);           
       end if;
-- If the test is finished, then write the temporary lines to the file.         
       if(finished = '1') then
           writeline(picture_output, L_R);
       end if;
    end if;
end process logging_ans;


-- There is no check_ans process in this testbench, since the param-module and the global concepts is 
-- validated while testing the parallell architecture.
-- The result to use later on is loged in the picture_output - file.

end architecture behavior; 